/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )

/*********************************************************/
/*                                                       */
/*********************************************************/
#if 0
int 
taucs_symccs_solve_llt(void* vL, double* x, double* b)
{
  taucs_ccs_matrix* L = (taucs_ccs_matrix*) vL;
  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  if (L->uplo != 'l') return -1;

  n = L->n;

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<n; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    /* 
    for (ip = (L->colptr)[j]; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      if (i==j) {
	Ajj = (L->values)[ip];
	break;
      }
    }
    */

    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }

    /*
    for (ip = (L->colptr)[j]; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      if (i != j) {
	Aij = (L->values)[ip];
	x[i] -= y[j]*Aij;
      }
    }
    */
  }
  
  /* Solve L^T x = y */

  for (i=n-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }
    /*
    for (jp = (L->colptr)[i]; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      if (i != j) {
	Aij = (L->values)[jp];
	y[i] -= x[j]*Aij;
      }
    }
    */

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    /*
    for (jp = (L->colptr)[i]; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      if (i==j) {
	Aii = (L->values)[jp];
	break;
      }
    }
    */

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}

/***************** SOLVE LLT PARTIAL ********************/

int 
taucs_symccs_solve_schur(symccs_matrix* L,
			 symccs_matrix* schur_comp,
			 void   (*schur_precond_fn)(void*,double* x,double* b),
			 void*  schur_precond_args,
			 int    maxits,
			 double convratio,
			 double* x, double* b)
{
  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  int p;

  if (L->uplo != 'l') return -1;

  n = L->n;
  p = n - (schur_comp->n);

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<p; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }
  }

  /* 
     now y_1 is computed, L_11 y_1 = b_1, 
     x_2 holds (b_2 - L_21 y_1).
     move y_2 <- x_2.
  */

  for (i=p; i<n; i++) y[i] = x[i];

  /* Now solve x_2 <- (A_22 - L_21 L_21^T)^-1 y_2 */ 
  
  /*taucs_printf("symccs_solve_schur: calling CG on Schur complement\n");*/
  taucs_conjugate_gradients (schur_comp,
			     schur_precond_fn,
			     schur_precond_args,
			     x+p,          /* this is x_2 */
			     y+p,          /* this is y_2 */
			     maxits,       /* itermax */
			     convratio     /* conv tolerance */
			     );
  /*taucs_printf("symccs_solve_llt_partial: CG on Schur complement returned\n");*/
  
  /* Now we have x_2, solve L_11^T x_1 = y_1 - L_21^T x_2 */

  for (i=p-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}
#endif
/*********************************************************/
/* CCS                                                   */
/*********************************************************/

int 
taucs_ccs_solve_llt(void* vL, double* x, double* b)
{
  taucs_ccs_matrix* L = (taucs_ccs_matrix*) vL;

  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  if (!(L->flags & TAUCS_TRIANGULAR)) {
    taucs_printf("taucs_ccs_solve_llt: factor matrix must be triangular\n");
    return -1;
  }
  if (!(L->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_solve_llt: lower part must be represented\n");
    return -1;
  }

  n = L->n;

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<n; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    /* 
    for (ip = (L->colptr)[j]; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      if (i==j) {
	Ajj = (L->values)[ip];
	break;
      }
    }
    */

    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }

    /*
    for (ip = (L->colptr)[j]; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      if (i != j) {
	Aij = (L->values)[ip];
	x[i] -= y[j]*Aij;
      }
    }
    */
  }
  
  /* Solve L^T x = y */

  for (i=n-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }
    /*
    for (jp = (L->colptr)[i]; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      if (i != j) {
	Aij = (L->values)[jp];
	y[i] -= x[j]*Aij;
      }
    }
    */

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    /*
    for (jp = (L->colptr)[i]; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      if (i==j) {
	Aii = (L->values)[jp];
	break;
      }
    }
    */

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}

/***************** SOLVE LLT PARTIAL ********************/

int 
taucs_ccs_solve_schur(taucs_ccs_matrix* L,
		      taucs_ccs_matrix* schur_comp,
		      int    (*schur_precond_fn)(void*,double* x,double* b),
		      void*  schur_precond_args,
		      int    maxits,
		      double convratio,
		      double* x, double* b)
{
  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  int p;

  if (!(L->flags & TAUCS_TRIANGULAR)) {
    taucs_printf("taucs_ccs_solve_llt: factor matrix must be triangular\n");
    return -1;
  }
  if (!(L->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_solve_llt: lower part must be represented\n");
    return -1;
  }

  n = L->n;
  p = n - (schur_comp->n);

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<p; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }
  }

  /* 
     now y_1 is computed, L_11 y_1 = b_1, 
     x_2 holds (b_2 - L_21 y_1).
     move y_2 <- x_2.
  */

  for (i=p; i<n; i++) y[i] = x[i];

  /* Now solve x_2 <- (A_22 - L_21 L_21^T)^-1 y_2 */ 
  
  /*taucs_printf("symccs_solve_schur: calling CG on Schur complement\n");*/
  taucs_conjugate_gradients (schur_comp,
			     schur_precond_fn,
			     schur_precond_args,
			     x+p,          /* this is x_2 */
			     y+p,          /* this is y_2 */
			     maxits,       /* itermax */
			     convratio     /* conv tolerance */
			     );
  /*taucs_printf("taucs_ccs_solve_llt_partial: CG on Schur complement returned\n");*/
  
  /* Now we have x_2, solve L_11^T x_1 = y_1 - L_21^T x_2 */

  for (i=p-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}

/*********************************************************/
/* OLD                                                   */
/*********************************************************/

#if 0
static int 
symccs_solve_llt_partial(symccs_matrix* L,
			 int p, 
			 double* x, double* b)
{
  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  symccs_matrix* SchurComp;
  symccs_matrix* F;         /* factored part of L */

  symccs_split(L,&F,&SchurComp,p);

  L = F;

  if (L->uplo != 'l') return -1;

  n = L->n;

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<p; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }
  }

  /* 
     now y_1 is computed, L_11 y_1 = b_1, 
     x_2 holds (b_2 - L_21 y_1).
     move y_2 <- x_2.
  */

  for (i=p; i<n; i++) y[i] = x[i];

  /* Now solve x_2 <- (A_22 - L_21 L_21^T)^-1 y_2 */ 
  
  taucs_printf("symccs_solve_llt_partial: calling CG on Schur complement\n");
  pcg (SchurComp,
       NULL,         /* no preconditioning */
       x+p,          /* this is x_2 */
       y+p,          /* this is y_2 */
       9999,         /* itermax */
       1e-12          /* conv tolerance */
       );
  taucs_printf("symccs_solve_llt_partial: CG on Schur complement returned\n");
  
  /* Now we have x_2, solve L_11^T x_1 = y_1 - L_21^T x_2 */

  for (i=p-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}

int symccs_solve_llt_partial_orig(symccs_matrix* L,
			     int p, 
			     double* x, double* b)
{
  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  symccs_matrix SchurComp;

  if (L->uplo != 'l') return -1;

  n = L->n;

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  for (j=0; j<p; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    y[j] = x[j] / Ajj;

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }
  }

  /* 
     now y_1 is computed, L_11 y_1 = b_1, 
     x_2 holds (b_2 - L_21 y_1).
     move y_2 <- x_2.
  */

  for (i=p; i<n; i++) y[i] = x[i];

  /* Now solve x_2 <- (A_22 - L_21 L_21^T)^-1 y_2 */ 
  
  SchurComp.uplo   = L->uplo;
  SchurComp.n      = n-p;
  SchurComp.colptr = (L->colptr) + p;
  SchurComp.rowind = L->rowind;
  SchurComp.indshift = p;
  SchurComp.values = L->values;

  taucs_printf("symccs_solve_llt_partial: calling CG on Schur complement\n");
  pcg (&SchurComp,
       NULL,         /* no preconditioning */
       x+p,          /* this is x_2 */
       y+p,          /* this is y_2 */
       9999,         /* itermax */
       1e-12          /* conv tolerance */
       );
  taucs_printf("symccs_solve_llt_partial: CG on Schur complement returned\n");
  
  /* Now we have x_2, solve L_11^T x_1 = y_1 - L_21^T x_2 */

  for (i=p-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = (L->values)[jp];

    x[i] = y[i] / Aii;

  }

  free(y);

  return 0;
}
#endif

/*********************************************************/
/* LDL^T solve                                           */
/*********************************************************/

int taucs_ccs_solve_ldlt(void* vL, double* x, double* b)
{
  taucs_ccs_matrix* L = (taucs_ccs_matrix*) vL;

  int n;
  int i,j;
  int ip,jp;
  double  Aij, Ajj, Aii;
  double* y;

  /* taucs_printf("taucs_ccs_solve_ldlt: starting\n"); */

  if (!(L->flags & TAUCS_TRIANGULAR)) {
    taucs_printf("taucs_ccs_solve_ldlt: factor matrix must be triangular\n");
    return -1;
  }
  if (!(L->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_solve_ldlt: lower part must be represented\n");
    return -1;
  }

  n = L->n;

  y = (double*) malloc( n * sizeof(double) );
  if (!y) return -1;

  for (i=0; i<n; i++) x[i] = b[i];

  /* Solve L y = b = x  */

  /* taucs_printf("taucs_ccs_solve_ldlt: solving L y = b\n"); */

  for (j=0; j<n; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = 1.0;
    
    y[j] = x[j] / Ajj;

    if (isnan(y[j]) || isinf(y[j])) {
      taucs_printf("taucs_ccs_solve_ldlt: inf/nan in column %d (L); %e/%e\n",
		 j,x[j],Ajj);
    }

    /*printf("A(%d,%d) = %lg; y[%d] = %lg\n",j,j,Ajj,i,y[i]);*/

    for (ip = (L->colptr)[j] + 1; ip < (L->colptr)[j+1]; ip++) {
      i = (L->rowind)[ip];
      Aij = (L->values)[ip];
      x[i] -= y[j]*Aij;
    }

  }
  
  /* Solve D y = y */

  for (j=0; j<n; j++) {

    /* we put diagonal elements first */
    ip = (L->colptr)[j];
    i = (L->rowind)[ip];
    assert (i==j);
    Ajj = (L->values)[ip];
    
    y[j] = y[j] / Ajj;
  }

  /* Solve L^T x = y */

  /* taucs_printf("taucs_ccs_solve_ldlt: solving L^T x = y\n");*/

  for (i=n-1; i>=0; i--) {

    for (jp = (L->colptr)[i]+1; jp < (L->colptr)[i+1]; jp++) {
      j = (L->rowind)[jp];
      Aij = (L->values)[jp];
      y[i] -= x[j]*Aij;
    }

    jp = (L->colptr)[i];
    j = (L->rowind)[jp];
    Aii = 1.0;

    x[i] = y[i] / Aii;

    if (isnan(x[i]) || isinf(x[i]))
	      taucs_printf("symccs_solve_ldlt: inf/nan in row %d (LT)\n",i);

    //printf("A(%d,%d) = %lg; x[%d] = %lg\n",i,i,Aii,i,x[i]);
  }

  free(y);

  /* taucs_printf("symccs_solve_ldlt: done\n");*/

  return 0;
}


/*********************************************************/
/*                                                       */
/*********************************************************/
